from typing import Iterable, Union, Tuple
import math
import numpy as np
from itertools import repeat
from functools import reduce
import operator

import poisson_disc

from ._imageloaders import ImageCollection, ImageBase


class PatchSampler(object):
    """Patch sampling algorithm to retrieve image patches/windows from images.

    If the image used for extracting patches has a mask associated to it, only
    patches from masked regions of the image are retrieved.

    Parameters
    ----------
    patch_size : Union[int, Iterable[int], dict]
        Size in pixels of the patches extracted on each axis. Only rectangular
        patches (hyper-cuboids) are supported by now. If a single int is
        passed, that size is used for all dimensions. If an iterable (list,
        tuple) is passed, each value will be assigned to the corresponding axes
         in `spatial_axes`, the size of `patch_size` must match the lenght of 
        `spatial_axes'. If a dict is passed, this should have at least the size
         of the patch of the axes listed in `spatial_axes`.  Use the same 
        convention as how Zarr structure array chunks in order to handle patch 
        shapes and channels correctly.
    min_area : Union[int, float]
        Minimum patch area covered by the mask to consider it samplable. A
        number in range [0, 1) will be used as percentage of the patch size. A
        number >= 1 will be considered as the number of minimum patch pixels
        covered by the mask.
    spatial_axes : str
        The spatial axes from where patches can be extracted.
    """
    def __init__(self, patch_size: Union[int, Iterable[int], dict],
                 min_area: Union[int, float] = 1,
                 spatial_axes: str ="ZYX"):
        # The maximum chunk sizes are used to generate a reference sampling
        # position array used fo every sampled chunk.
        self._max_chunk_size = dict((ax, 0) for ax in spatial_axes)

        if isinstance(patch_size, (list, tuple)):
            if len(patch_size) != len(spatial_axes):
                raise ValueError(f"The size of `patch_size` must match the "
                                 f"number of axes in `spatial_axes`, got "
                                 f"{len(patch_size)} for {spatial_axes}")

            patch_size = dict((ax, ps)
                              for ax, ps in zip(spatial_axes, patch_size))

        elif isinstance(patch_size, int):
            patch_size = dict((ax, patch_size) for ax in spatial_axes)

        elif not isinstance(patch_size, dict):
            raise ValueError(f"Patch size must be a dictionary specifying the"
                             f" patch size of each axes, an iterable (list, "
                             f"tuple) with the same order as the spatial axes,"
                             f" or an integer for a cubic patch. Received "
                             f"{patch_size} of type {type(patch_size)}")

        self.spatial_axes = spatial_axes

        self._patch_size = dict(
            (ax, patch_size.get(ax, 1))
            for ax in spatial_axes
        )

        self._min_area = min_area

    def _compute_corners(self, non_zero_pos: tuple, axes: str) -> np.ndarray:
        toplefts = np.stack(non_zero_pos).T
        toplefts = toplefts.astype(np.float32)

        toplefts_corners = []

        dim = len(axes)
        factors = 2 ** np.arange(dim + 1)
        for d in range(2 ** dim):
            corner_value =  np.array((d % factors[1:]) // factors[:-1],
                                     dtype=np.float32)
            toplefts_corners.append(
                toplefts + (1 - 1e-4) * corner_value
            )

        corners = np.stack(toplefts_corners)

        return corners

    def _compute_overlap(self, corners: np.ndarray, shape: np.ndarray,
                         ref_shape: np.ndarray) -> Tuple[np.ndarray,
                                                         np.ndarray]:
        scaled_corners = corners * shape[None, None]
        tls_scaled = scaled_corners / ref_shape[None, None]
        tls_idx = np.floor(tls_scaled)

        corners_cut = np.maximum(tls_scaled[0], tls_idx[-1])

        dist2cut = np.fabs(scaled_corners - corners_cut[None])
        coverage = np.prod(dist2cut, axis=-1)

        return coverage, tls_idx.astype(np.int64)

    def _compute_grid(self, chunk_mask: np.ndarray,
                      mask_axes: str,
                      mask_scale: dict,
                      patch_size: dict,
                      image_size: dict):

        mask_relative_shape = np.array(
            [1 / m_scl
             for ax, m_scl in mask_scale.items()
             if ax in self.spatial_axes
            ],
            dtype=np.float32
        )

        patch_shape = np.array(
            [patch_size[ax]
             for ax in mask_axes
             if ax in self.spatial_axes
            ],
            dtype=np.float32
        )

        # If the patch sizes are greater than the relative shape of the mask
        # with respect to the input image, use the mask coordinates as
        # reference to overlap the coordinates of the sampling patches.
        # Otherwise, use the patches coordinates instead.
        if all(map(operator.ge, patch_shape, mask_relative_shape)):
            active_coordinates = np.nonzero(chunk_mask)
            ref_axes = mask_axes

            ref_shape = patch_shape
            shape = mask_relative_shape

            mask_is_greater = False
            
            patch_ratio = [
                    image_size[ax] // ps
                    for ax, ps in zip(mask_axes, patch_shape.astype(np.int64))
                    if ax in self.spatial_axes
            ]

            if not all(patch_ratio):
                return np.empty(
                    [0] * len(set(mask_axes).intersection(self.spatial_axes)),
                    dtype=np.int64
                )

        else:
            active_coordinates = np.meshgrid(
                *[np.arange(image_size[ax] // ps)
                  for ax, ps in zip(mask_axes, patch_shape)
                  if ax in self.spatial_axes]
            )

            active_coordinates = tuple(
                coord_ax.flatten()
                for coord_ax in active_coordinates
            )

            ref_axes = "".join([
                ax for ax in self.spatial_axes if ax in mask_axes
            ])

            ref_shape = mask_relative_shape
            shape = patch_shape

            mask_is_greater = True

        corners = self._compute_corners(active_coordinates, axes=ref_axes)

        (coverage,
         corners_idx)= self._compute_overlap(corners, shape, ref_shape)

        if mask_is_greater:
            # The mask ratio is greater than the patches size
            mask_values = chunk_mask[tuple(corners_idx.T)].T
            patches_coverage = coverage * mask_values

            covered_tls = corners[0, ...].astype(np.int64)

        else:
            # The mask ratio is less than the patches size
            patch_coordinates = np.ravel_multi_index(tuple(corners_idx.T),
                                                     chunk_mask.shape)
            patches_coverage = np.bincount(patch_coordinates.flatten(),
                                           weights=coverage.flatten())
            patches_coverage = np.take(patches_coverage, patch_coordinates).T

            covered_tls = corners_idx[0, ...]

        patches_coverage = np.sum(patches_coverage, axis=0)

        # Compute minimum area covered by masked regions to sample a patch.
        min_area = self._min_area
        if min_area < 1:
            min_area *= patch_shape.prod()

        minumum_covered_tls = covered_tls[patches_coverage > min_area]

        if not mask_is_greater:
            # Collapse to unique coordinates since there will be multiple
            # instances of the same patch.
            minumum_covered_tls = np.ravel_multi_index(
                tuple(minumum_covered_tls.T),
                patch_ratio,
                mode="clip"
            )

            minumum_covered_tls = np.unique(minumum_covered_tls)

            minumum_covered_tls = np.unravel_index(
                minumum_covered_tls,
                patch_ratio
            )

            minumum_covered_tls = np.stack(minumum_covered_tls).T

        return minumum_covered_tls * patch_shape[None].astype(np.int64)

    def _compute_valid_toplefts(self, chunk_mask: np.ndarray, mask_axes: str,
                                mask_scale: dict,
                                patch_size: dict,
                                image_size: dict):
        return self._compute_grid(chunk_mask, mask_axes, mask_scale,
                                  patch_size,
                                  image_size)

    def _compute_toplefts_slices(self, mask: ImageBase, image_shape: dict,
                                 patch_size: dict,
                                 valid_mask_toplefts: np.ndarray,
                                 chunk_tlbr: dict):
        toplefts = []
        for tls in valid_mask_toplefts:
            curr_tl = []

            for ax in self.spatial_axes:
                if ax in mask.axes:
                    tl = ((chunk_tlbr[ax].start
                           if chunk_tlbr[ax].start is not None else 0)
                          + tls[mask.axes.index(ax)])
                    br = ((chunk_tlbr[ax].start
                           if chunk_tlbr[ax].start is not None else 0)
                          + tls[mask.axes.index(ax)] + patch_size[ax])
                    if br <= image_shape[ax]:
                        curr_tl.append((ax, slice(tl, br)))
                    else:
                        break
                else:
                    curr_tl.append((ax, slice(0, 1)))

            else:
                toplefts.append(dict(curr_tl))

        return toplefts

    def compute_chunks(self,
                       image_collection: ImageCollection) -> Iterable[dict]:
        """Compute the top-left positions of the chunks in which the image is
        divided before taking patches from them.

        The chunk size should match the zarr's chunk sizes to make the patch
        extraction more efficient.

        Parameters
        ----------
        image : zarrdataset.ImageCollection
            The image from where the patches are extracted.

        Returns
        -------
        chunks_grids : list of tuples of slices
            Each valid chunk is returned in form of a slice that can be
            extracted from the image. The slices are stored as tuples, with one
            slice per axis.
        """
        # Use the mask as base to determine the valid patches that can be
        # retrieved from the image.
        image = image_collection.collection[image_collection.reference_mode]
        mask = image_collection.collection[image_collection.mask_mode]

        spatial_chunk_sizes = dict(
            (ax, chk)
            for ax, chk in zip(image.axes, image.chunk_size)
            if ax in self.spatial_axes
        )

        image_shape = dict(map(tuple, zip(image.axes, image.shape)))

        self._max_chunk_size = dict(
            (ax, (min(max(self._max_chunk_size[ax],
                          spatial_chunk_sizes[ax],
                          self._patch_size[ax]),
                      image_shape[ax]))
                 if ax in image.axes else 1)
            for ax in self.spatial_axes
        )

        chunk_tlbr = dict(
            map(tuple, zip(self.spatial_axes, repeat(slice(None))))
        )

        chunk_mask = mask[chunk_tlbr]

        valid_mask_toplefts = self._compute_grid(
            chunk_mask,
            mask.axes,
            mask.scale,
            self._max_chunk_size,
            image_shape
        )

        chunks_slices = self._compute_toplefts_slices(
            mask,
            image_shape=image_shape,
            patch_size=self._max_chunk_size,
            valid_mask_toplefts=valid_mask_toplefts,
            chunk_tlbr=chunk_tlbr
        )

        return chunks_slices

    def compute_patches(self, image_collection: ImageCollection,
                        chunk_tlbr: dict) -> Iterable[dict]:
        image = image_collection.collection[image_collection.reference_mode]
        mask = image_collection.collection[image_collection.mask_mode]
        image_shape = dict(map(tuple, zip(image.axes, image.shape)))
        chunk_size = dict(
            (ax,
             (ctb.stop if ctb.stop is not None else image_shape[ax])
             - (ctb.start if ctb.start is not None else 0)
            )
            for ax, ctb in chunk_tlbr.items()
        )

        chunk_mask = mask[chunk_tlbr]

        valid_mask_toplefts = self._compute_valid_toplefts(
            chunk_mask,
            mask.axes,
            mask.scale,
            self._patch_size,
            chunk_size
        )

        patches_slices = self._compute_toplefts_slices(
            mask,
            image_shape=image_shape,
            patch_size=self._patch_size,
            valid_mask_toplefts=valid_mask_toplefts,
            chunk_tlbr=chunk_tlbr
        )

        return patches_slices

    def __repr__(self) -> str:
        """String representation of classes derived from PatchSampler.
        """
        return (f"{type(self)} for sampling patches of size "
                f"{self._patch_size}.")


class BlueNoisePatchSampler(PatchSampler):
    """Patch sampler that retrieves patches from coordinates sampled using the
    Bridson sampling algorithm, also known as Blue-noise sampling algorithm.

    Parameters
    ----------
    patch_size : int, iterable, dict
        Size in pixels of the patches extracted. Only squared patches are
        supported by now.
    resample_positions : bool
        Whether positions should be resampled for each new chunk or not.
    allow_overlap : bool
        Whether overlapping of patches is allowed or not.
    """
    def __init__(self, patch_size: Union[int, Iterable[int], dict],
                 resample_positions=False,
                 allow_overlap=False,
                 **kwargs):
        super(BlueNoisePatchSampler, self).__init__(patch_size)
        self._base_chunk_tls = None
        self._resample_positions = resample_positions
        self._allow_overlap = allow_overlap
 
    def compute_sampling_positions(self, force=False) -> None:
        """Compute the sampling positions using blue-noise sampling.

        Parameters
        ----------
        force: bool
            Whether force resampling positions, or reuse existing positions.
        """
        if self._base_chunk_tls is not None and not force:
            return

        radius = max(self._patch_size.values())
        if self._allow_overlap:
            radius *= math.sqrt(2)

        samplable_axes = [ax for ax in self.spatial_axes
                          if self._max_chunk_size[ax] > radius]
        dims = [self._max_chunk_size[ax] - radius for ax in samplable_axes]

        if len(samplable_axes):
            base_chunk_tls = poisson_disc.Bridson_sampling(dims=np.array(dims),
                                                           radius=radius,
                                                           k=30)
            base_chunk_tls = base_chunk_tls.astype(np.int64)

            base_chunk_tls = [
                base_chunk_tls[:, samplable_axes.index(ax)]
                if ax in samplable_axes else np.zeros(len(base_chunk_tls),
                                                      dtype=np.int64)
                for ax in self.spatial_axes
            ]
            self._base_chunk_tls = np.stack(base_chunk_tls).T

        else:
            self._base_chunk_tls = np.zeros((1, len(self.spatial_axes)),
                                            dtype=np.int64)

    def _compute_valid_toplefts(self,
                                chunk_mask: np.ndarray,
                                mask_axes: str,
                                mask_scale: dict,
                                patch_size: dict,
                                image_shape: dict):
        self.compute_sampling_positions(force=self._resample_positions)

        # Filter sampling positions that does not contain any mask portion.
        sampling_pos = np.hstack(
            tuple(
                self._base_chunk_tls[:, self.spatial_axes.index(ax), None]
                if ax in self.spatial_axes else
                np.zeros((len(self._base_chunk_tls), 1), dtype=np.float32)
                for ax in mask_axes
            )
        )

        spatial_patch_sizes = np.array([
            patch_size[ax]
            for ax in mask_axes
            if ax in self.spatial_axes
        ])

        mask_corners = self._compute_corners(np.nonzero(chunk_mask),
                                             mask_axes)

        dist = (mask_corners[None, :, :, :]
                - sampling_pos[:, None, None, :].astype(np.float32)
                - spatial_patch_sizes[None, None, None, :] / 2)

        mask_samplable_pos, = np.nonzero(
            np.any(
                np.all(dist < spatial_patch_sizes[None, None, None], axis=-1),
                axis=(1, 2)
            )
        )

        toplefts = sampling_pos[mask_samplable_pos]

        return toplefts



from typing import Union, Iterable, Callable

import math
import zarr
import numpy as np

import PIL
from PIL import Image
from io import BytesIO

from ._utils import (map_axes_order, connect_s3, scale_coords, select_axes,
                     parse_rois,
                     translate2roi)

try:
    import tifffile
    TIFFFILE_SUPPORT = True

except ModuleNotFoundError:
    TIFFFILE_SUPPORT = False


def image2array(arr_src: Union[str, zarr.Group, zarr.Array, np.ndarray],
                data_group: Union[str, int, None] = None,
                zarr_store: Union[zarr.storage.Store, None] = None):
    """Open images stored in zarr format or any image format that can be opened
    by PIL as an array.

    Parameters
    ----------
    arr_src : Union[str, zarr.Group, zarr.Array, np.ndarray]
        The image filename, or zarr object, to be loaded as a zarr array.
    data_group : Union[str, int, None]
        The group within the zarr file from where the array is loaded. This is
        used only when the input file is a zarr object.
    zarr_store : Union[zarr.storage.Store, None]
        The class used to open the zarr file. Leave it as None to let this
        function to use the most suitable depending to the data location
        (s3/remote: FSStore, local disk: DirectoryStore).

    Returns
    -------
    arr : zarr.Array
        The image as a zarr array.
    store: None, zarr.storage.Store, PIL.Image, tifffile.ZarrTiffStore
        A connection to the image file that is kept open until the dataset is
        not needed, so this connection can be closed properly.
    """
    if isinstance(arr_src, zarr.Group):
        if data_group is not None and len(data_group):
            arr = arr_src[data_group]
        else:
            raise ValueError(f"Cannot use {arr_src} because it is a Zarr Group"
                             f" and a Zarr Array was expected. Please specify"
                             f" the group where the array is stored.")
        return arr, None

    elif isinstance(arr_src, zarr.Array):
        return arr_src, None

    elif isinstance(arr_src, str) and ".zarr" in arr_src:
        s3_obj = connect_s3(arr_src)

        if zarr_store is None:
            # If zarr_store is not set by the user, assign the most suitable
            # according to the image location (remote: FSStore,
            # local: DirectoryStore).
            if s3_obj is not None:
                zarr_store = zarr.storage.FSStore
            else:
                zarr_store = zarr.storage.DirectoryStore

        store = zarr_store(arr_src)
        grp = zarr.open(store, mode="r")

        if isinstance(grp, zarr.Group):
            if data_group is not None and len(data_group):
                arr = grp[data_group]
            else:
                raise ValueError(f"Cannot use {arr_src} because it is a Zarr "
                                 f"Group and the group where the array is "
                                 f"stored was not specified.")
        else:
            arr = grp

        return arr, None

    elif isinstance(arr_src, np.ndarray):
        arr = zarr.array(data=arr_src, shape=arr_src.shape,
                         chunks=arr_src.shape)
        return arr, None
    
    # Try to create a connection with the file, to determine if it is a remote
    # resource or local file.
    s3_obj = connect_s3(arr_src)

    if TIFFFILE_SUPPORT and s3_obj is None:
        # Try to open the input file with tifffile (if installed).
        try:
            if (data_group is None
              or (isinstance(data_group, str) and not len(data_group))):
                tiff_args = dict(
                    key=None,
                    level=None,
                    series=None
                )
            elif isinstance(data_group, str) and len(data_group):
                data_group_split = data_group.split("/")

                if len(data_group_split) > 1:
                    tiff_args = dict(
                        key=int(data_group_split[0]),
                        level=int(data_group_split[1]),
                        series=None
                    )
                else:
                    tiff_args = dict(
                        key=int(data_group_split[0]),
                        level=None,
                        series=None
                    )

            elif isinstance(data_group, int):
                tiff_args = dict(
                    key=data_group,
                    level=None,
                    series=None
                )
            else:
                raise ValueError(f"Incorrect data group format "
                                 f"{type(data_group)}")

            store = tifffile.imread(arr_src, aszarr=True, **tiff_args)
            arr = zarr.open(store, mode="r")

            return arr, store

        except tifffile.tifffile.TiffFileError:
            pass

    # If the input is a path to an image stored in a format
    # supported by PIL, open it and use it as a numpy array.
    try:
        if s3_obj is not None:
            # The image is stored in a S3 bucket
            filename = arr_src.split(s3_obj["endpoint_url"]
                                     + "/"
                                     + s3_obj["bucket_name"])[1][1:]
            im_bytes = s3_obj["s3"].get_object(
                Bucket=s3_obj["bucket_name"],
                Key=filename)["Body"].read()
            store = Image.open(BytesIO(im_bytes))

        else:
            # The image is stored locally
            store = Image.open(arr_src, mode="r")

        channels = len(store.getbands())
        height = store.size[1]
        width = store.size[0]

        arr_shape = [height, width] + ([channels] if channels > 1 else [])
        arr = zarr.array(data=np.array(store),
                         shape=arr_shape,
                         chunks=arr_shape,
                         dtype=np.uint8)
        return arr, store

    except PIL.UnidentifiedImageError:
        pass

    raise ValueError(f"The file/object {arr_src} cannot be opened by Zarr "
                     f"{', TiffFile, ' if TIFFFILE_SUPPORT else ''} or PIL")


class ImageBase(object):
    arr = None
    spatial_axes = "ZYX"
    source_axes = None
    axes = None
    mode = ""
    permute_order = None
    _store = None
    _new_axes = ""
    _drop_axes = ""
    _scale = None
    _shape = None
    _spatial_reference_shape = None
    _spatial_reference_axes = None
    _chunk_size = None
    _cached_coords = None
    _image_func = None

    def __init__(self, shape: Iterable[int],
                 chunk_size: Union[Iterable[int], None]=None,
                 source_axes: str="",
                 mode: str=""):
        if chunk_size is None:
            chunk_size = shape

        self.source_axes = source_axes
        self.axes = source_axes
        self.permute_order = list(range(len(source_axes)))
        self.arr = zarr.ones(shape=shape, dtype=bool, chunks=chunk_size)
        self.roi = tuple([slice(None)] * len(source_axes))
        self.mode = mode
        self._chunk_size = chunk_size

    def _iscached(self, coords):
        if self._cached_coords is None:
            return False
        else:
            return all(
                map(lambda coord, cache, s:
                    (cache.start if cache.start is not None else 0)
                    <= (coord.start if coord.start is not None else 0)
                    and (coord.stop if coord.stop is not None else s)
                    <= (cache.stop if cache.stop is not None else s),
                    coords,
                    self._cached_coords,
                    self.arr.shape))

    def _cache_chunk(self, index):
        if not self._iscached(index):
            self._cached_coords = tuple(
                map(lambda i, chk, s:
                    slice(chk * int(i.start / chk)
                          if i.start is not None else 0,
                          min(s, chk * int(math.ceil(i.stop / chk)))
                          if i.stop is not None else None,
                          None),
                    index,
                    self.arr.chunks,
                    self.arr.shape)
            )
            self._cache = self.arr[self._cached_coords]

        cached_index = tuple(
            map(lambda cache, i:
                slice((i.start - cache.start) if i.start is not None else 0,
                      (i.stop - cache.start) if i.stop is not None else None,
                      None),
                self._cached_coords, index)
        )

        return cached_index

    def __getitem__(self, index : Union[slice, tuple, dict]) -> np.ndarray:
        if self._spatial_reference_axes is not None:
            spatial_reference_axes = self._spatial_reference_axes
        else:
            spatial_reference_axes = self.axes

        if isinstance(index, slice):
            index = [index] * len(spatial_reference_axes)

        if not isinstance(index, dict):
            # Arrange the indices requested using the reference image axes
            # ordering.
            index = dict(
                ((ax, sel)
                for ax, sel in zip(spatial_reference_axes, index))
            )

        mode_index, _ = select_axes(self.axes, index)
        mode_scales = tuple(self.scale[ax] for ax in self.axes)

        mode_index = scale_coords(mode_index, mode_scales)

        mode_index = dict(
            ((ax, sel)
            for ax, sel in zip(self.axes, mode_index))
        )

        # Locate the mode_index within the ROI:
        roi_mode_index = translate2roi(mode_index, self.roi, self.source_axes,
                                       self.axes)

        # Save the corresponding cache of this patch for faster access.
        cached_index = self._cache_chunk(roi_mode_index)
        selection = self._cache[cached_index]

        # Permute the axes order to match `axes`
        selection = selection.transpose(self.permute_order)

        # Drop axes with length 1 that are not in `axes`.
        out_shape = [s
                     for s, p_a in zip(selection.shape, self.permute_order)
                     if self.source_axes[p_a] not in self._drop_axes]
        selection = selection.reshape(out_shape)

        # Add axes requested in `axes` that does not exist on `source_axes`.
        selection = np.expand_dims(selection, tuple(self.axes.index(a)
                                                    for a in self._new_axes))

        if self._image_func is not None:
            selection = self._image_func(selection)

        return selection

    def _compute_shapes(self):
        if self._shape is not None:
            return

        self._shape = []
        self._chunk_size = []

        for a in self.axes:
            if a in self.source_axes:
                a_i = self.source_axes.index(a)
                r = self.roi[a_i]
                s = self.arr.shape[a_i]
                c = min(self.arr.chunks[a_i], s)

                r_start = 0 if r.start is None else r.start
                r_stop = s if r.stop is None else r.stop

                self._shape.append(r_stop - r_start)
                self._chunk_size.append(c)

            else:
                self._shape.append(1)
                self._chunk_size.append(1)

        if self._spatial_reference_axes is None:
            self._spatial_reference_axes = [
                ax for ax in self.axes if ax in self.spatial_axes
                ]

        if self._spatial_reference_shape is None:
            self._spatial_reference_shape = [
                s for s, ax in zip(self._shape, self.axes)
                if ax in self.spatial_axes
                ]

    def rescale(self, spatial_reference_shape: Union[Iterable[int], None]=None,
                spatial_reference_axes: Union[str, None]=None) -> None:
        """Rescale this image using the `spatial_reference_shape` as reference.

        Parameters
        ----------
        spatial_reference_shape: Union[Iterable[int], None]
            Reference image shape used to match extracted regions from this
            image (e.g., when calling __getitem__, or ImageBase[slice(...)])
        spatial_reference_axes: Union[str, None]
            Rescale only this axes from the image, keeping the rest unscaled.
        """
        if (self._scale is not None
           and spatial_reference_shape is None
           and spatial_reference_axes is None):
            return

        if spatial_reference_shape is not None:
            self._spatial_reference_shape = spatial_reference_shape

        if spatial_reference_axes is not None:
            self._spatial_reference_axes = spatial_reference_axes

        self._compute_shapes()

        self._scale = {}

        for ax, s in zip(self.axes, self.shape):
            if ax in self._spatial_reference_axes:
                ax_i = self._spatial_reference_axes.index(ax)
                ref_s = self._spatial_reference_shape[ax_i]

                self._scale[ax] = s / ref_s
            else:
                self._scale[ax] = 1.0

    @property
    def shape(self) -> Iterable[int]:
        self._compute_shapes()
        return self._shape

    @property
    def chunk_size(self) -> Iterable[int]:
        self._compute_shapes()
        return self._chunk_size

    @property
    def scale(self) -> dict:
        self.rescale()
        return self._scale


class ImageLoader(ImageBase):
    """Image lazy loader class.

    Opens the zarr file, or any image that can be open by TiffFile or PIL, as a
    Zarr array.

    Parameters
    ----------
    filename: str
    source_axes: str
    data_group: Union[str, None]
    axes: Union[str, None]
    roi: Union[str, slice, Iterable[slice], None]
    image_func: Union[Callable, None]
    zarr_store: Union[zarr.storage.Store, None]
    spatial_axes: str
    mode: str
    """
    def __init__(self, filename: str, source_axes: str,
                 data_group: Union[str, None] = None,
                 axes: Union[str, None] = None,
                 roi: Union[str, slice, Iterable[slice], None] = None,
                 image_func: Union[Callable, None] = None,
                 zarr_store: Union[zarr.storage.Store, None] = None,
                 spatial_axes: str = "ZYX",
                 mode: str = ""):
        self.mode = mode
        self.spatial_axes = spatial_axes

        if roi is None:
            parsed_roi = [slice(None)] * len(source_axes)
        elif isinstance(roi, str):
            parsed_roi = parse_rois([roi])[0]
        elif isinstance(roi, (list, tuple)):
            if len(roi) != len(source_axes):
                raise ValueError(f"ROIs does not match the number of the array"
                                 f" axes. Expected {len(source_axes)}, got "
                                 f"{len(roi)}")
            elif not all([isinstance(roi_ax, slice) for roi_ax in roi]):
                raise ValueError(f"ROIs must be slices, but got "
                                 f"{[type(roi_ax) for roi_ax in roi]}")
            else:
                parsed_roi = roi
        elif isinstance(roi, slice):
            if (len(source_axes) > 1
              and not (roi.start is None and roi.stop is None)):
                raise ValueError(f"ROIs must specify a slice per axes. "
                                 f"Expected {len(source_axes)} slices, got "
                                 f"only {roi}")
            else:
                parsed_roi = [roi] * len(source_axes)
        else:
            raise ValueError(f"Incorrect ROI format, expected a list of "
                             f"slices, or a parsable string, got {roi}")

        roi_slices = list(
            map(lambda r:
                slice(r.start if r.start is not None else 0, r.stop, None),
                parsed_roi)
        )

        (self.arr,
         self._store) = image2array(filename, data_group=data_group,
                                    zarr_store=zarr_store)

        self.roi = roi_slices
        self.source_axes = source_axes
        self.axes = source_axes

        if axes is not None and axes != source_axes:
            source_axes_list = list(source_axes)
            self._drop_axes = list(set(source_axes) - set(axes))
            for d_a in sorted((source_axes_list.index(a)
                               for a in self._drop_axes),
                              reverse=True):
                if self.roi[d_a].stop is not None:
                    roi_len = self.roi[d_a].stop
                else:
                    roi_len = self.arr.shape[d_a]

                roi_len -= self.roi[d_a].start

                if roi_len > 1:
                    raise ValueError(f"Cannot drop axis `{source_axes[d_a]}` "
                                     f"(from `source_axes={source_axes}`) "
                                     f"because no selection was made for it, "
                                     f"and it is not being used as image axes "
                                     f"thereafter (`axes={axes}`).")

            self.permute_order = map_axes_order(source_axes, axes)
            self._new_axes = list(set(axes) - set(source_axes))
            self.axes = axes
        else:
            self.permute_order = list(range(len(self.axes)))

        self._image_func = image_func

        if image_func is not None:
            self.axes = image_func.axes

    def __del__(self):
        # Close the connection to the image file
        if self._store is not None:
            self._store.close()


class ImageCollection(object):
    """A class to contain a collection of inputs from different modalities.

    This is used to match images with their respective labels and masks.

    Parameters
    ----------
    collection_args : dict
        Collection arguments containing specifications to open `images`,
        `masks`, `labels`, etc.
    spatial_axes : str
        Spatial axes of the dataset, which are used to match different
        modalities using as reference these axes from the `images` collection.
    """
    def __init__(self, collection_args : dict,
                 spatial_axes: str = "ZYX"):

        self.reference_mode = list(collection_args.keys())[0]

        self.spatial_axes = spatial_axes

        self.collection = dict((
            (mode, ImageLoader(spatial_axes=spatial_axes, mode=mode,
                               **mode_args))
            for mode, mode_args in collection_args.items()
        ))

        self._generate_mask()
        self.reset_scales()

    def _generate_mask(self):
        mask_modes = list(
            filter(lambda k: "mask" in k,
                   self.collection.keys()
            )
        )

        if len(mask_modes) > 0:
            self.mask_mode = mask_modes[0]
            return

        ref_axes = self.collection[self.reference_mode].axes
        ref_shape = self.collection[self.reference_mode].shape
        ref_chunk_size = self.collection[self.reference_mode].chunk_size

        mask_axes = list(set(self.spatial_axes).intersection(ref_axes))
        mask_axes_ord = map_axes_order(mask_axes, ref_axes)
        mask_axes = [mask_axes[a] for a in mask_axes_ord]

        mask_chunk_size = [
            int(math.ceil(s / c))
            for s, c, a in zip(ref_shape, ref_chunk_size, ref_axes)
            if a in mask_axes
            ]

        self.collection["masks"] = ImageBase(shape=mask_chunk_size,
                                             chunk_size=mask_chunk_size,
                                             source_axes=mask_axes,
                                             mode="masks")
        self.mask_mode = "masks"

    def reset_scales(self) -> None:
        """Reset the scales between data modalities to match the `images`
        collection shape on the `spatial_axes` only.
        """
        img_shape = self.collection[self.reference_mode].shape
        img_source_axes = self.collection[self.reference_mode].source_axes
        img_axes = self.collection[self.reference_mode].axes

        spatial_reference_axes = [
            ax
            for ax in img_axes if ax in self.spatial_axes
        ]
        spatial_reference_shape = [
            img_shape[img_axes.index(ax)]
            if ax in img_source_axes else 1
            for ax in spatial_reference_axes
        ]

        for img in self.collection.values():
            img.rescale(spatial_reference_shape, spatial_reference_axes)

    def __getitem__(self, index):
        collection_set = dict(
            (mode, img[index])
            for mode, img in self.collection.items()
        )

        return collection_set



from typing import Union
import numpy as np
from skimage import morphology, color, filters, transform


class MaskGenerator(object):
    """Base class to define transformation functions on ImageBase objects.
    """
    def __init__(self, axes):
        self.axes = axes

    def _compute_transform(self, image: np.ndarray) -> np.ndarray:
        raise NotImplementedError("This is a virtual method and has to be "
                                  "implemented by a derived image "
                                  "transformation class")

    def __call__(self, image: np.ndarray) -> np.ndarray:
        return self._compute_transform(image)


class WSITissueMaskGenerator(MaskGenerator):
    """Mask generator for tissue objects in Whole Slide Images (WSI).

    This will compute a mask of scale `mask_scale` from the input image where
    tissue (darker pixels than white background) are present.

    Parameters
    ----------
    mask_scale : float
    min_size : int
    area_threshold : int,
    thresh : Union[float, None]
    axes : str
    """
    def __init__(self, mask_scale : float = 1/16, min_size : int = 16,
                 area_threshold : int = 128,
                 thresh : Union[float, None] = None,
                 axes : str = "YX"):
        super(WSITissueMaskGenerator, self).__init__(axes=axes)

        self._mask_scale = mask_scale
        self._min_size = min_size
        self._area_threshold_2 = area_threshold ** 2
        self._thresh = thresh

    def _compute_transform(self, image: np.ndarray) -> np.ndarray:
        gray = color.rgb2gray(image)
        scaled_gray = transform.rescale(gray, scale=self._mask_scale, order=0,
                                        preserve_range=True)

        if self._thresh is None:
            self._thresh = filters.threshold_otsu(scaled_gray)

        chunk_mask = scaled_gray < self._thresh
        chunk_mask = morphology.remove_small_objects(
            chunk_mask == 0, min_size=self._min_size ** 2, connectivity=2)
        chunk_mask = morphology.remove_small_holes(
            chunk_mask, area_threshold=self._area_threshold_2)
        chunk_mask = morphology.binary_erosion(
            chunk_mask, morphology.disk(max(1, self._min_size // 2)))
        mask = morphology.binary_dilation(
            chunk_mask, morphology.disk(self._min_size))

        if "Z" in self.axes:
            mask = mask[None, ...]

        return mask



from typing import Iterable, Union, List
import os
import math
from itertools import repeat
from urllib.parse import urlparse
import requests
import boto3
import zarr


def parse_rois(rois_str: Iterable[str]) -> List[slice]:
    """Parse a list of strings defining ROIs.

    Parameters
    ----------
    rois_str : Iterable[str]
        A list of regions of interest parsed from the input string.

    Returns
    -------
    rois : List[slice]
        A list of slices to select regions from arrays.
    """
    rois = []

    if isinstance(rois_str, str):
        rois_str = [rois_str]

    # Create a tuple of slices to define each ROI.
    for roi in rois_str:
        if ":" in roi:
            start_coords, axis_lengths = roi.split(":")

            start_coords = tuple([int(c.strip("\n\r ()"))
                                  for c in start_coords.split(",")])

            axis_lengths = tuple([int(ln.strip("\n\r ()"))
                                  for ln in axis_lengths.split(",")])

            roi_slices = []
            for c_i, l_i in zip(start_coords, axis_lengths):
                if l_i > 0:
                    roi_slices.append(slice(c_i, c_i + l_i, None))
                elif c_i > 0:
                    roi_slices.append(slice(c_i, None, None))
                else:
                    roi_slices.append(slice(None))

            roi_slices = tuple(roi_slices)
        else:
            roi_slices = slice(None)

        rois.append(roi_slices)

    return rois


def parse_metadata(filename: str, default_source_axes: str,
                   default_data_group: Union[str, int, None]=None,
                   default_axes: Union[str, None]=None,
                   default_rois: Union[Iterable[slice], None]=None,
                   override_meta: bool=False) -> List[dict]:
    """Parse the filename, data groups, axes ordering, ROIs from `filename`.

    The different fields must be separated by a semicolon (;).
    After parsed the filename, data group, and axes, any number of ROIs are
    accepted; however, this parser only supports rectangle shaped ROIs.

    Parameters
    ----------
    filename : str
        Path to the image.
    default_source_axes : src
        Default source axes ordering used when no axes are present in the
        filename string.
    default_data_group : Union[str, int, None]
        Default data group used when data group is present in the filename
        string.
    default_axes : Union[str, None]
        Default output axes ordering used when no target axes are present in
        the filename string.
    default_rois : Union[Iterable[slice], None]
        Default roi used when no rois are present in the filename string.
    override_meta : bool
        Whether to override the values parsed from the filename with the
        defaults or not.

    Returns
    -------
    parsed_metadata : list of dicts
        A list of parsed values from the filename. A single element is
        retrieved for each ROI in the filename.

    Notes
    -----
    The parsed metadata dictionary has the following structure.
    fn : str
        The parsed filename
    data_group: str or None
        The group for zarr images, or key for tiff files
    source_axes: str or None
        The orignal axes ordering
    axes: str or None
        The axes ordering as it being used from the array (may involve
        permuting, dropping unused axes, and creating new axes)
    rois : list of slices
        A list of regions of interest parsed from the input string.

    Data groups are expected for files with hierarchical structure, like zarr
    and tiff files. This is usually found on pyramid-like files, where the main
    image is stored in group `0/0`, and any downsampling version of that are
    stored in subgroups `0/1', '0/2`, etc. If the file does not have any
    hierarchical structure, leave the space in blank.

    Original axes and permuted desired order are expected as
    `source axes order`:`target axes order`. Axes identifiers can appear only
    once. Different number of target axes from the original number of axes can
    be requested. For more axes, dummy axes will be added in the position of
    the new requested axes. For less axes, the undefined axes will be
    positioned arbitrarly before those specified axes. It is responsability of
    the user to make remove unused axes with the a corresponding ROI.
    If no reordering is required, do not add the colon (:) on that section.

    ROIs are spected as (start coordinate):(lenght of the ROI). Coordinates and
    lenghts are expected for each one of the axes in the image, since all ROIs
    are taken before any axes reordering. Negative values can be used to select
    all indices of that axis.

    Example 1:
    test_file.zarr;0/0;TCZYX:YXC;(0,0,0,0,0):(1,-1,1,-1,-1)

    This will parse a ROI from `test_file.zarr` from the group `0/0`, which is
    expetect to have axes ordering as Time (T), Channels (C), Depth (Z),
    Height (Y), and Width (X), into an axes order of Height, Width and
    Channels. Index `0` is selected from both Time (T) and Depth(Z) axes.

    Example 2:
    test_file.zarr;1;CYX;(0,0,0):(-1,4,3);(0,5,5):(-1,2,2)

    This will parse a ROI from `test_file.zarr` from the array at group `1`,
    which is expetect to have axes ordering as Channels (C), Height (Y), and
    Width (X). From that array, two ROIs are extracted, one from coordinates
    (0, 0) with a shape (4, 3), and other from coordinates (5, 5) with shape
    (2, 2). Both ROIs use all the available channels.

    Example 3:
    test_array.zarr;;CYX:YXC

    The `test_array.zarr` file stores an array without any hierarchy, so no
    data group is required. Only the channels axes is required to be moved from
    the first position to the last one. Because no ROIs are defined, the full
    the array will be used.
    """
    source_axes = None
    target_axes = None
    data_group = None
    rois_str = []

    if default_axes is None:
        default_axes = default_source_axes

    if isinstance(filename, str):
        # There can be filenames with `;` on them, so take the dot separating
        # the filename and extension as reference to start spliting the
        # filename.
        fn_base_split = filename.split(".")
        ext_meta = fn_base_split[-1].split(";")
        fn_ext = ext_meta[0]
        fn = ".".join(fn_base_split[:-1]) + "." + fn_ext

        if len(ext_meta) > 1:
            data_group = ext_meta[1]
            axes_str = ext_meta[2].split(":")
            rois_str = ext_meta[3:]

            # Parse source and target axes ordering
            source_axes = axes_str[0]
            if len(axes_str) > 1:
                target_axes = axes_str[1]
    else:
        fn = filename

    if not source_axes or (override_meta and default_source_axes is not None):
        source_axes = default_source_axes

    if not data_group or (override_meta and default_data_group is not None):
        data_group = default_data_group

    if not target_axes or (override_meta and default_axes is not None):
        target_axes = default_axes

    rois = parse_rois(rois_str)
    if not rois or (override_meta and default_rois is not None):
        if not isinstance(default_rois, list):
            default_rois = [default_rois]
        rois = default_rois

    parsed_metadata = [
        {"filename": fn,
         "data_group": data_group,
         "source_axes": source_axes,
         "axes": target_axes,
         "roi": roi}
        for roi in rois
    ]

    return parsed_metadata


def map_axes_order(source_axes: str, target_axes: str="YX"):
    """Get the indices of a set of axes that reorders it to match another set
    of axes.

    Parameters
    ----------
    source_axes : str
        Set of axes to be reordered to match `target_axes`.
    target_axes : str
        Set of axes in the desired order.

    Returns
    -------
    transpose_order : list of ints
        The indices order in `source_axes` that makes it match `target_axes`.
        If `source_axes` has more axes than `target_axes`, the unspecified axes
        will be moved to the front of the ordering, leaving the `target_axes`
        at the trailing positions.

    Notes
    -----
    The axes must be passed as a string in format `XYZ`, and cannot be appear
    more than once.
    """
    # Take only existing axes of the image
    target_axes = [t_ax for t_ax in target_axes if t_ax in source_axes]
    unused_axes = list(set(source_axes) - set(target_axes))
    transpose_order = [source_axes.index(a)
                       for a in unused_axes + list(target_axes)]

    return transpose_order


def select_axes(source_axes : str, axes_selection : dict):
    """Get a sliced selection of axes from a zarr array.

    Parameters
    ----------
    source_axes : str
        Ordered set of axes on the original array.
    axes_selection : dict
        A relationship of what index to take from each specified axes.
        Unspecified axes are fully retrieved.

    Returns
    -------
    sel_slices : tuple of slices
        A tuple of slices that can be used to select the indices of the
        specified axes from an array.
    unfixed_axes : str
        A string with the ordered set of remaining axes of the array.

    Notes
    -----
    The axes must be passed as a string in format `XYZ`, and cannot be appear
    more than once.
    """
    sel_slices = []
    unfixed_axes = list(source_axes)

    for ax in source_axes:
        idx = axes_selection.get(ax, None)

        if idx is None or (idx.start is None and idx.stop is None):
            sel_slices.append(slice(None))

        else:
            # Remove this axis from the set of axes that the array has after
            # fixing this to index `idx`.
            unfixed_axes.remove(ax)
            sel_slices.append(idx)

    # These are the new set of source_axes of the array after fixing some axes
    # to the specified indices.
    unfixed_axes = "".join(unfixed_axes)
    sel_slices = tuple(sel_slices)

    return sel_slices, unfixed_axes


def connect_s3(filename_sample: str):
    """Stablish a connection with a S3 bucket.

    Parameters
    ----------
    filename_sample : str
        A sample filename containing the S3 end-point and bucket names.

    Returns
    -------
    s3_obj : dict of None
        A dictionary containing the bucket name and a boto3 client connected to
        the S3 bucket. If the filename does not points to a S3 bucket, this
        returns None.
    """
    if (isinstance(filename_sample, str)
        and (filename_sample.startswith("s3")
             or filename_sample.startswith("http"))):
        protocol = filename_sample.split(":")[0]
        endpoint = protocol + "://" + urlparse(filename_sample).netloc
        s3_obj = dict(
            bucket_name=filename_sample.split(endpoint)[1].split("/")[1],
            endpoint_url=endpoint,
            s3=boto3.client("s3", aws_access_key_id="",
                            aws_secret_access_key="",
                            region_name="us-east-2",
                            endpoint_url=endpoint))

        s3_obj["s3"]._request_signer.sign = (lambda *args, **kwargs: None)
    else:
        s3_obj = None

    return s3_obj


def isconsolidated(arr_src: Union[str, zarr.Group, zarr.Array]):
    """Check if the zarr file is consolidated so it is faster to open.

    Parameters
    ----------
    arr_src : Union[str, zarr.Group, zarr.Array]
        The image filename, or zarr object, to be checked.

    Returns
    -------
    is_consolidated : bool
        Whether zarr array in `arr_src` is consolidated or not.
    """
    is_consolidated = False

    # TODO: Add more protocols to check, like gcp and s3
    if "http" in arr_src or "https" in arr_src:
        response = requests.get(arr_src + "/.zmetadata")
        is_consolidated = response.status_code == 200
    else:
        is_consolidated = os.path.exists(os.path.join(arr_src, ".zmetadata"))

    return is_consolidated


def scale_coords(selection_range : Iterable[slice],
                 scale : Union[float, Iterable[float]] = 1.0):
    """Scale a set of top-lefts, bottom-rights coordinates, in any dimension,
    by `scale` factor.

    Parameters
    ----------
    selection_range : Iterable[slice]
        The selection range from an n-dimensional array to scale. This can be a
        range (start, end), single index, or None, defining the range of
        indices taken from each axis.
    scale : Union[float, Iterable[float]]
        The factor to rescale the selection range of each axes. If a single
        value is passed, all axes are rescaled by that factor.

    Returns
    -------
    scaled_selection_range : tuple of slices
        The scaled selection range as a tuple of slices
    """
    scaled_selection_range = []
    if not isinstance(scale, (list, tuple)):
        scale = repeat(scale)

    for ax_range, s in zip(selection_range, scale):
        if isinstance(ax_range, (slice, range)):
            if ax_range.start is None or ax_range.stop is None:
                scaled_selection_range.append(
                    slice(None)
                )
            else:
                scaled_selection_range.append(
                    slice(int(ax_range.start * s),
                          int(math.ceil(ax_range.stop * s)),
                          ax_range.step)
                )
        elif isinstance(ax_range, (tuple, list)):
            scaled_selection_range.append(
                slice(int(ax_range[0] * s), int(math.ceil(ax_range[1] * s)),
                      None)
            )
        elif isinstance(ax_range, int):
            scaled_selection_range.append(
                slice(int(ax_range * s), int(math.ceil((ax_range + 1) * s)),
                      None)
            )
        elif ax_range is None:
            scaled_selection_range.append(
                slice(None)
            )
        else:
            raise ValueError(f"Axis selection {ax_range} is not supported. "
                             f"Only ranges in form [start, end], "
                             f"single index, and None, are supported.")

    scaled_selection_range = tuple(scaled_selection_range)
    return scaled_selection_range


def translate2roi(index : dict, roi : tuple, source_axes : str, axes : str):
    roi_mode_index = {}
    for a_i, ax in enumerate(source_axes):
        r = roi[a_i]

        if ax in axes and ax in index:
            i = index[ax]

            i_start = i.start if i.start is not None else 0
            r_start = r.start if r.start is not None else 0

            i_start = i_start + r_start
            i_stop = (r_start + i.stop) if i.stop is not None else None

            if i_stop is not None and r.stop is not None:
                i_stop = min(i_stop, r.stop)
            elif i_stop is None:
                i_stop = r.stop

            if i_stop is None and i_start == 0:
                i_start = None

            roi_mode_index[ax] = slice(i_start, i_stop, None)
        else:
            roi_mode_index[ax] = r

    roi_mode_index, _ = select_axes(source_axes, roi_mode_index)

    return roi_mode_index



from collections import OrderedDict
from typing import Iterable, Union, Callable
from functools import reduce, partial
import operator
import random
import zarr
import numpy as np

from ._utils import parse_metadata
from ._imageloaders import ImageCollection
from ._samplers import PatchSampler

try:
    from tqdm import tqdm
    TQDM_SUPPORT=True

except ModuleNotFoundError:
    TQDM_SUPPORT=False

    # This removes the dependency on tqdm when it is not installed
    class tqdm(object):
        """Placeholder used to remove dependency in `tqdm`, so ZarrDataset can
        still be used when `tqdm` is not installed.
        """
        def __init__(self, *args, **kwargs):
            pass

        def update(self, *args, **kwargs):
            pass

        def set_description(self, *args, **kwargs):
            pass

        def close(self, *args, **kwargs):
            pass


try:
    import torch
    from torch.utils.data import IterableDataset
    PYTORCH_SUPPORT=True

except ModuleNotFoundError:
    import logging
    logging.warning('PyTorch is not installed, the BaseZarrDataset class will '
                    'still work as a python iterator')
    IterableDataset = object
    PYTORCH_SUPPORT=False


def zarrdataset_worker_init_fn(worker_id):
    """ZarrDataset multithread workers initialization function.
    """
    if not PYTORCH_SUPPORT:
        return

    worker_info = torch.utils.data.get_worker_info()
    w_sel = slice(worker_id, None, worker_info.num_workers)

    dataset_obj = worker_info.dataset

    # Reset the random number generators in each worker.
    torch_seed = torch.initial_seed()
    random.seed(torch_seed)
    np.random.seed(torch_seed % (2**32 - 1))

    dataset_obj._worker_sel = w_sel
    dataset_obj._worker_id = worker_id
    dataset_obj._num_workers = worker_info.num_workers

def chained_zarrdataset_worker_init_fn(worker_id):
    """ZarrDataset multithread workers initialization function for PyTorch's
    ChainedDatasets.
    """
    if not PYTORCH_SUPPORT:
        return

    worker_info = torch.utils.data.get_worker_info()
    dataset_obj = worker_info.dataset

    # Reset the random number generators in each worker.
    torch_seed = torch.initial_seed()
    random.seed(torch_seed)
    np.random.seed(torch_seed % (2**32 - 1))

    # Open a copy of the dataset on each worker.
    dataset_obj.datasets = \
        dataset_obj.datasets[worker_id::worker_info.num_workers]

    for ds in dataset_obj.datasets:
        ds._worker_id = worker_id
        ds._num_workers = 1


class ImageSample():
    _current_patch_idx = 0
    _ordering = None
    _rng_seed = None
    num_patches = None

    def __init__(self, im_id: int, chk_id: int, shuffle: bool = False):
        self.im_id = im_id
        self.chk_id = chk_id
        self._shuffle = shuffle

        if self._shuffle:
            self._rng_seed = random.randint(1, 100000)

    def free_sampler(self):
        del self._ordering
        self._ordering = None

    def next_patch(self):
        if self._shuffle and self._ordering is None:
            curr_state = random.getstate()
            random.seed(self._rng_seed)
            self._ordering = list(range(self.num_patches))
            random.shuffle(self._ordering)
            random.setstate(curr_state)

        if self._shuffle:
            curr_patch = self._ordering[self._current_patch_idx]
        else:
            curr_patch = self._current_patch_idx

        self._current_patch_idx += 1
        is_empty = self._current_patch_idx >= self.num_patches

        return curr_patch, is_empty


class DatasetSpecs(dict):
    """Data specification guidelines to add image modalities to a ZarrDataset.

    Parameters
    ----------
    modality: str
        Specifies the use of this dataset (input image data, labels, masks).
    filenames: Union[str,
                     Iterable[str],
                     zarr.Group,
                     Iterable[zarr.Group],
                     zarr.Array,
                     Iterable[zarr.Array],
                     np.ndarray,
                     Iterable[np.ndarray]]
        The input source either specified by a path/url to a file or a
        supported array-like object, or list of them.
    source_axes: str
        The orignal array axes ordering.
    axes: Union[str, None]
        The axes ordering as it being used from the array (may involve
        permuting, dropping unused axes, and creating new axes).
    data_group: Union[str, int, None]
        The group for zarr images, or key for tiff files
    roi: Union[str, slice, Iterable[slice], None]
        Regions of interest from the input array that can be used for data
        sampling.
    image_loader_func: Union[Callable, None]
        A transformation applied to the input array before sampling. Could be
        used to define a mask generation function. This is not a data
        augmentation transform. To specify a data augmetation transform use
        `transform` instead.
    zarr_store: Union[zarr.storage.Store, None]
        A specific zarr.storage.Store class to be used to load zarr files.
    transform: Union[Callable, None]
        A transform applied to the array before returning it after sampling.
        This can be used to specify data augmentation transforms.
    add_to_output: bool
        Whether add this modality to the output after sampling or not. For
        example, labels would be added to the output along with the input
        image array, while masks might not be needed.
    """
    def __init__(self, modality: str,
                 filenames: Union[str, Iterable[str], zarr.Group,
                                  Iterable[zarr.Group],
                                  zarr.Array,
                                  Iterable[zarr.Array],
                                  np.ndarray,
                                  Iterable[np.ndarray]],
                 source_axes: str,
                 axes: Union[str, None] = None,
                 data_group: Union[str, int, None] = None,
                 roi: Union[str, slice, Iterable[slice], None] = None,
                 image_loader_func: Union[Callable, None] = None,
                 zarr_store: Union[zarr.storage.Store, None] = None,
                 transform: Union[Callable, None] = None,
                 add_to_output: bool = True):

        super().__init__()

        self["modality"] = modality
        self["filenames"] = filenames
        self["source_axes"] = source_axes
        self["axes"] = axes
        self["data_group"] = data_group
        self["roi"] = roi
        self["image_loader_func"] = image_loader_func
        self["transforms"] = OrderedDict()
        self["zarr_store"] = zarr_store
        self["add_to_output"] = add_to_output

        if transform is not None:
            self["transforms"][(modality, )] = transform


class ImagesDatasetSpecs(DatasetSpecs):
    """Specification to add `image` data to a ZarrDataset.

    Parameters
    ----------
    filenames: Union[str,
                     Iterable[str],
                     zarr.Group,
                     Iterable[zarr.Group],
                     zarr.Array,
                     Iterable[zarr.Array],
                     np.ndarray,
                     Iterable[np.ndarray]]
        The input source either specified by a path/url to a file or a
        supported array-like object, or list of them.
    source_axes: str
        The orignal array axes ordering.
    axes: Union[str, None]
        The axes ordering as it being used from the array (may involve
        permuting, dropping unused axes, and creating new axes).
    data_group: Union[str, int, None]
        The group for zarr images, or key for tiff files
    roi: Union[str, slice, Iterable[slice], None]
        Regions of interest from the input array that can be used for data
        sampling.
    image_loader_func: Union[Callable, None]
        A transformation applied to the input array before sampling. Could be
        used to define a mask generation function. This is not a data
        augmentation transform. To specify a data augmetation transform use
        `transform` instead.
    zarr_store: Union[zarr.storage.Store, None]
        A specific zarr.storage.Store class to be used to load zarr files.
    transform: Union[Callable, None]
        A transform applied to the array before returning it after sampling.
        This can be used to specify data augmentation transforms.
    modality: str
        Specifies the use of this dataset (default is `images` for image data).
    """
    def __init__(self,
                 filenames: Union[str, Iterable[str], zarr.Group,
                                  Iterable[zarr.Group],
                                  zarr.Array,
                                  Iterable[zarr.Array],
                                  np.ndarray,
                                  Iterable[np.ndarray]],
                 source_axes: str,
                 axes: Union[str, None] = None,
                 data_group: Union[str, int, None] = None,
                 roi: Union[str, slice, Iterable[slice], None] = None,
                 image_loader_func: Union[Callable, None] = None,
                 zarr_store: Union[zarr.storage.Store, None] = None,
                 transform: Union[Callable, None] = None,
                 modality: str ="images"):

        super().__init__(
            modality,
            filenames,
            source_axes,
            axes,
            data_group,
            roi,
            image_loader_func,
            zarr_store,
            transform,
            add_to_output=True
        )


class LabelsDatasetSpecs(DatasetSpecs):
    """Specification to add `labels` to a ZarrDataset.

    Parameters
    ----------
    filenames: Union[str,
                     Iterable[str],
                     zarr.Group,
                     Iterable[zarr.Group],
                     zarr.Array,
                     Iterable[zarr.Array],
                     np.ndarray,
                     Iterable[np.ndarray]]
        The input source either specified by a path/url to a file or a
        supported array-like object, or list of them.
    source_axes: str
        The orignal array axes ordering.
    axes: Union[str, None]
        The axes ordering as it being used from the array (may involve
        permuting, dropping unused axes, and creating new axes).
    data_group: Union[str, int, None]
        The group for zarr images, or key for tiff files
    roi: Union[str, slice, Iterable[slice], None]
        Regions of interest from the input array that can be used for data
        sampling.
    image_loader_func: Union[Callable, None]
        A transformation applied to the input array before sampling. Could be
        used to define a mask generation function. This is not a data
        augmentation transform. To specify a data augmetation transform use
        `transform` instead.
    zarr_store: Union[zarr.storage.Store, None]
        A specific zarr.storage.Store class to be used to load zarr files.
    transform: Union[Callable, None]
        A transform applied to the array before returning it after sampling.
        This can be used to specify data augmentation transforms.
    input_label_transform: Union[Callable, None]
        A transform applied to the array before returning it after sampling.
        This can be used to specify data augmentation transforms.
    modality: str
        Specifies the use of this dataset (default is `labels`).
    """
    def __init__(self,
                 filenames: Union[str, Iterable[str], zarr.Group,
                                  Iterable[zarr.Group],
                                  zarr.Array,
                                  Iterable[zarr.Array],
                                  np.ndarray,
                                  Iterable[np.ndarray]],
                 source_axes: str,
                 axes: Union[str, None] = None,
                 data_group: Union[str, int, None] = None,
                 roi: Union[str, slice, Iterable[slice], None] = None,
                 image_loader_func: Union[Callable, None] = None,
                 zarr_store: Union[zarr.storage.Store, None] = None,
                 transform: Union[Callable, None] = None,
                 input_label_transform: Union[Callable, None] = None,
                 input_mode: str = "images",
                 modality: str = "labels"):

        super().__init__(
            modality,
            filenames,
            source_axes,
            axes,
            data_group,
            roi,
            image_loader_func,
            zarr_store,
            transform,
            add_to_output=True
        )

        if input_label_transform is not None:
            self["transforms"][("labels", )] = transform
            self["transforms"][(input_mode, "labels")] =\
                input_label_transform


class MasksDatasetSpecs(DatasetSpecs):
    """Specification to add `masks` to a ZarrDataset.

    Parameters
    ----------
    filenames: Union[str,
                     Iterable[str],
                     zarr.Group,
                     Iterable[zarr.Group],
                     zarr.Array,
                     Iterable[zarr.Array],
                     np.ndarray,
                     Iterable[np.ndarray]]
        The input source either specified by a path/url to a file or a
        supported array-like object, or list of them.
    source_axes: str
        The orignal array axes ordering.
    axes: Union[str, None]
        The axes ordering as it being used from the array (may involve
        permuting, dropping unused axes, and creating new axes).
    data_group: Union[str, int, None]
        The group for zarr images, or key for tiff files
    roi: Union[str, slice, Iterable[slice], None]
        Regions of interest from the input array that can be used for data
        sampling.
    image_loader_func: Union[Callable, None]
        A transformation applied to the input array before sampling. Could be
        used to define a mask generation function. This is not a data
        augmentation transform. To specify a data augmetation transform use
        `transform` instead.
    zarr_store: Union[zarr.storage.Store, None]
        A specific zarr.storage.Store class to be used to load zarr files.
    modality: str
        Specifies the use of this dataset (default is `masks`).
    """
    def __init__(self,
                 filenames: Union[str, Iterable[str], zarr.Group,
                                  Iterable[zarr.Group],
                                  zarr.Array,
                                  Iterable[zarr.Array],
                                  np.ndarray,
                                  Iterable[np.ndarray]],
                 source_axes: str,
                 axes: Union[str, None] = None,
                 data_group: Union[str, int, None] = None,
                 roi: Union[str, slice, Iterable[slice], None] = None,
                 image_loader_func: Union[Callable, None] = None,
                 zarr_store: Union[zarr.storage.Store, None] = None,
                 modality: str = "masks"):

        super().__init__(
            modality,
            filenames,
            source_axes,
            axes,
            data_group,
            roi,
            image_loader_func,
            zarr_store,
            None,
            add_to_output=False
        )


class ZarrDataset(IterableDataset):
    """A Zarr-based dataset class capable of handling large volumes of image
    data stored in OME-NGFF Zarr format. This class can match the coordinates
    of the different image modalities to those in the `images` mode, so labels
    and masks are retrieved from these same coordinates. All spatial axes are
    scaled using the `images` mode as reference, therefore labels and masks do
    not need to share the same sizes as the arrays in the `images` mode.

    Parameters
    ----------
    dataset_specs: Union[dict, Iterable[dict], None]
        A list of dictionaries containing the specifications of the datasets
        used as inputs.
    patch_sampler: Union[PatchSampler, None]
        The patch sampling algorithm used to extract patches from images.
    shuffle: bool
        Whether samples are extracted in order or at random.
    progress_bar: bool
        Display a progress bar to show the status of data initialization.
        Requires `tqdm` to be installed in the environment.
    return_positions: bool
        Return the top-left positions from where the samples where extracted
        along with the set of patches.
    return_worker_id: bool
        Return the worker id that extracted the sample.
    draw_same_chunk: bool
        Whether continue extracting samples from the same chunk, until
        depleting the posible patches to extract, before extract samples from
        a different chunk. This can be used to reduce the overhead of
        retrieving different chunks when sampling patches at random locations
        whithin the input image.
    """
    def __init__(self, dataset_specs: Union[Iterable[dict], None] = None,
                 patch_sampler: Union[PatchSampler, None] = None,
                 shuffle: bool = False,
                 progress_bar: bool = False,
                 return_positions: bool = False,
                 return_worker_id: bool = False,
                 draw_same_chunk: bool = False):

        self._worker_sel = slice(None)
        self._worker_id = 0
        self._num_workers = 1

        self._shuffle = shuffle
        self._progress_bar = progress_bar
        self._return_positions = return_positions
        self._return_worker_id = return_worker_id
        self._draw_same_chunk = draw_same_chunk

        self._patch_sampler = patch_sampler

        self._transforms = OrderedDict()
        self._output_order = []

        self._collections = {}
        self._zarr_store = {}
        self._image_loader_func = {}

        self._arr_lists = []
        self._toplefts = []

        self._ref_mod = None

        self._initialized = False

        if dataset_specs is not None:
            # Iterate the modalities in the dataset specifications
            if not isinstance(dataset_specs, list):
                dataset_specs = [dataset_specs]

            for specs in dataset_specs:
                self.add_modality(**specs)

    def _initialize(self, force=False):
        if self._ref_mod is None:
            raise ValueError("No image modalities have been added to the "
                             "dataset. Use `.add_modality` to add at least one"
                             " modality to this dataset prior to "
                             "initialization.")

        if self._initialized and not force:
            return

        arr_lists = []
        toplefts = []

        if self._progress_bar:
            q = tqdm(desc="Preloading zarr files",
                     total=len(self._collections[self._ref_mod]),
                     position=self._worker_id)

        modes = self._collections.keys()

        for collection in zip(*self._collections.values()):
            collection = dict([(m, c) for m, c in zip(modes, collection)])
            for mode in collection.keys():
                collection[mode]["zarr_store"] = self._zarr_store[mode]
                collection[mode]["image_func"] = self._image_loader_func[mode]

            if self._progress_bar:
                q.set_description(f"Preloading image "
                                  f"{collection['images']['filename']}")

            curr_img = ImageCollection(collection)

            # If a patch sampler was passed, it is used to determine the
            # top-left and bottom-right coordinates of the valid samples that
            # can be drawn from images.
            if self._patch_sampler is not None:
                toplefts.append(self._patch_sampler.compute_chunks(curr_img))
            else:
                toplefts.append([
                    dict((ax, slice(None))
                         for ax in curr_img.collection[self._ref_mod].axes)
                    ]
                )

            arr_lists.append(curr_img)

            if self._progress_bar:
                q.update()

        if self._progress_bar:
            q.close()

        if len(arr_lists) == 1:
            self._arr_lists = np.array(arr_lists, dtype=object)
            self._toplefts = np.array([toplefts[0][self._worker_sel]],
                                      dtype=object)

        elif len(arr_lists) < self._num_workers:
            self._arr_lists = np.array(arr_lists, dtype=object)
            self._toplefts = np.array([tls[self._worker_sel]
                                       for tls in toplefts],
                                      dtype=object)

        else:
            self._arr_lists = np.array(arr_lists[self._worker_sel],
                                       dtype=object)
            self._toplefts = np.array(toplefts[self._worker_sel],
                                      dtype=object)

        self._initialized = True

    def __getitem__(self, tlbr):
        coords = tlbr if tlbr is not None else slice(None)
        patches = self._curr_collection[coords]

        for inputs, transform in self._transforms.items():
            res = transform(*tuple(patches[mode] for mode in inputs))

            if not isinstance(res, tuple):
                res = (res, )

            for mode, mode_res in zip(inputs, res):
                patches[mode] = mode_res

        patches = [patches[mode] for mode in self._output_order]

        return patches

    def __iter__(self):
        # Preload the files and masks associated with them
        self._initialize()

        samples = [
            ImageSample(im_id, chk_id, shuffle=self._shuffle)
            for im_id in range(len(self._arr_lists))
            for chk_id in range(len(self._toplefts[im_id]))
        ]

        # Shuffle chunks here if samples will come from the same chunk until
        # they are depleted.
        if self._shuffle and self._draw_same_chunk:
            random.shuffle(samples)

        prev_im_id = -1
        prev_chk_id = -1
        prev_chk = -1
        curr_chk = 0
        self._curr_collection = None

        while samples:
            # Shuffle chunks here if samples can come from different chunks.
            if self._shuffle and not self._draw_same_chunk:
                curr_chk = random.randrange(0, len(samples))

            im_id = samples[curr_chk].im_id
            chk_id = samples[curr_chk].chk_id

            chunk_tlbr = self._toplefts[im_id][chk_id]

            # If this sample is from a different image or chunk, free the
            # previous sample and re-sample the patches from the current chunk.
            if prev_im_id != im_id or chk_id != prev_chk_id:
                if prev_chk >= 0:
                    # Free the patch ordering from the previous chunk to save
                    # memory.
                    samples[prev_chk].free_sampler()

                prev_chk = curr_chk
                prev_chk_id = chk_id

                if prev_im_id != im_id:
                    prev_im_id = im_id
                    self._curr_collection = self._arr_lists[im_id]

                if self._patch_sampler is not None:
                    patches_tls = self._patch_sampler.compute_patches(
                        self._curr_collection,
                        chunk_tlbr
                    )

                else:
                    patches_tls = [chunk_tlbr]

                samples[curr_chk].num_patches = len(patches_tls)

                if not len(patches_tls):
                    samples.pop(curr_chk)
                    prev_chk = -1
                    continue

            # # Initialize the count of top-left positions for patches inside
            # # this chunk.
            # if samples[curr_chk].num_patches is None:

            curr_patch, is_empty = samples[curr_chk].next_patch()

            # When all possible patches have been extracted from the current
            # chunk, remove that chunk from the list of samples.
            if is_empty:
                samples.pop(curr_chk)
                prev_chk = -1

            patch_tlbr = patches_tls[curr_patch]
            patches = self.__getitem__(patch_tlbr)

            if self._return_positions:
                pos = [
                    [patch_tlbr[ax].start
                     if patch_tlbr[ax].start is not None else 0,
                     patch_tlbr[ax].stop
                     if patch_tlbr[ax].stop is not None else -1
                    ] if ax in patch_tlbr else [0, -1]
                    for ax in self._collections[self._ref_mod][0]["axes"]
                ]
                patches = [np.array(pos, dtype=np.int64)] + patches

            if self._return_worker_id:
                wid = [np.array(self._worker_id, dtype=np.int64)]
                patches = wid + patches

            if len(patches) > 1:
                patches = tuple(patches)
            else:
                patches = patches[0]

            yield patches

    def add_transform(self, modalities: Union[str, Iterable[str]],
                      transform: Callable):
        """Add a pre-processing transform pipeline to the dataset, applied to
        the arrays from modalities specified with `modes`. This will be
        performed after any other pre-processing transforms already registered.

        Parameters
        ----------
        modalities: Union[str, Iterable[str]]
            The modalities on which this transform is applied (e.g., ``images``
            to apply only on image arrays, or (``images``, ``labels``) to apply
            it to both, images and labels arrays)
        transform: Callable
            A function that receives the same number of inputs as specified in
            `modalities`, and returns that same number of outputs.
        """
        if isinstance(modalities, str):
            modalities = (modalities, )

        self._transforms[modalities] = transform

    def add_modality(self,
                     modality: str,
                     filenames: Union[str, Iterable[str], zarr.Group,
                                  Iterable[zarr.Group],
                                  zarr.Array,
                                  Iterable[zarr.Array],
                                  np.ndarray,
                                  Iterable[np.ndarray]],
                     source_axes: str,
                     axes: Union[str, None] = None,
                     data_group: Union[str, int, None] = None,
                     roi: Union[str, slice, Iterable[slice], None] = None,
                     image_loader_func: Union[Callable, None] = None,
                     zarr_store: Union[zarr.storage.Store, None] = None,
                     transforms: Union[OrderedDict, None] = None,
                     add_to_output: bool = True):
        """Add a new modality to the dataset.

        Parameters
        ----------
        modality: str
            The name of the new modality added (e.g., ``images``, ``labels``,
            ``masks``, etc.).
        filenames: Union[str,
                        Iterable[str],
                        zarr.Group,
                        Iterable[zarr.Group],
                        zarr.Array,
                        Iterable[zarr.Array],
                        np.ndarray,
                        Iterable[np.ndarray]]
            The input source either specified by a path/url to a file or a
            supported array-like object, or list of them.
        source_axes: str
            The orignal array axes ordering.
        axes: Union[str, None]
            The axes ordering as it being used from the array (may involve
            permuting, dropping unused axes, and creating new axes).
        data_group: Union[str, int, None]
            The group for zarr images, or key for tiff files
        roi: Union[str, slice, Iterable[slice], None]
            Regions of interest from the input array that can be used for data
            sampling.
        image_loader_func: Union[Callable, None]
            A transformation applied to the input array before sampling. Could
            be used to define a mask generation function. This is not a data
            augmentation transform. To specify a data augmetation transform use
            `transform` instead.
        zarr_store: Union[zarr.storage.Store, None]
            A specific zarr.storage.Store class to be used to load zarr files.
        transform: Union[OrderedDict, None]
            A list of transforms applied to arrays before yielding them. This
            can be used to specify data augmentation transforms, and can be
            applied to this and other existing modalities in this dataset.
            For example, to add a transform affecting images and labels, use
            the tuple ('images', 'labels') as key for that transform, and make
            sure the function associated to that key receives, and returns the
            same number of inputs and ouputs as specified in the key.
        add_to_output: bool
            Whether add this modality to the output after sampling or not. For
            example, labels would be added to the output along with the input
            image array, while masks might not be needed.
        """
        if self._ref_mod is None:
            self._ref_mod = modality

        self._collections[modality] = reduce(
            operator.add,
            map(partial(parse_metadata,
                        default_source_axes=source_axes,
                        default_data_group=data_group,
                        default_axes=axes,
                        default_rois=roi),
                filenames if isinstance(filenames, list) else [filenames]),
            []
        )

        if transforms is not None:
            for t_ord, t in transforms.items():
                self.add_transform(modalities=t_ord, transform=t)

        if add_to_output:
            self._output_order.append(modality)

        self._zarr_store[modality] = zarr_store

        self._image_loader_func[modality] = image_loader_func

    def __repr__(self) -> str:
        """ZarrDataset string representation.
        """
        repr_str = (f"ZarrDataset (PyTorch support:{PYTORCH_SUPPORT}, tqdm "
                    f"support :{TQDM_SUPPORT})"
                    + "\n"
                    + f"Modalities: {','.join(self._collections.keys())}"
                    + "\n"
                    + f"Transforms order: {list(self._transforms.keys())}"
                    + "\n"
                    + f"Using {self._ref_mod} modality as reference.")

        if self._patch_sampler is not None:
            repr_str += ("\n"
                         + f"Using {str(self._patch_sampler)}")

        return repr_str



from ._samplers import (PatchSampler,
                        BlueNoisePatchSampler)

from ._maskfuncs import (MaskGenerator,
                         WSITissueMaskGenerator)

from ._imageloaders import (image2array,
                            ImageBase,
                            ImageLoader,
                            ImageCollection)

from ._utils import (parse_rois,
                     parse_metadata,
                     map_axes_order,
                     select_axes,
                     connect_s3,
                     isconsolidated,
                     scale_coords,
                     translate2roi)

from ._augs import ToDtype

from ._zarrdataset import (DatasetSpecs,
                           ImagesDatasetSpecs,
                           LabelsDatasetSpecs,
                           MasksDatasetSpecs,
                           ZarrDataset,
                           zarrdataset_worker_init_fn,
                           chained_zarrdataset_worker_init_fn)


__all__ = ['DatasetSpecs',
           'ImagesDatasetSpecs',
           'LabelsDatasetSpecs',
           'MasksDatasetSpecs',
           'ZarrDataset',
           'zarrdataset_worker_init_fn',
           'chained_zarrdataset_worker_init_fn',
           'PatchSampler',
           'BlueNoisePatchSampler',
           'ToDtype',
           'parse_rois',
           'parse_metadata',
           'map_axes_order',
           'select_axes',
           'connect_s3',
           'isconsolidated',
           'scale_coords',
           'translate2roi',
           'image2array',
           'ImageBase',
           'ImageLoader',
           'ImageCollection',
           'MaskGenerator',
           'WSITissueMaskGenerator']



import numpy as np


class ToDtype(object):
    """Converted a numpy NDArray to the specified data type.

    Parameters
    ----------
    dtype : numpy.dtype
        The data type to cast the input array.
    """
    def __init__(self, dtype: np.dtype) -> None:
        self._dtype = dtype

    def __call__(self, image:np.ndarray) -> np.ndarray:
        """Casts the type of `image` to the data type specified with `dtype`.

        Parameters
        ----------
        image : np.ndarray
            A numpy NDArray

        Returns
        -------
        casted_image : numpy.ndarray
            The sampe input image
        """
        casted_image = image.astype(self._dtype)
        return casted_image

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dtype={self._dtype})"



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zarrdataset",
    version=os.environ.get('VERSION', '0.0.0'),
    maintainer="Fernando Cervantes",
    maintainer_email="fernando.cervantes@jax.org",
    description="Zarr-based dataset for PyTorch training pipelines. Written "
                "and maintained by the Research IT team at The Jackson "
                "Laboratory.",

    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheJacksonLaboratory/zarrdataset",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests>=2.31.0',
        'aiohttp>=3.7.4.post0,<4.0.0',
        'botocore>=1.31.17,<1.31.18',
        'boto3>=1.28.17,<1.28.18',
        'fsspec>=2023.9.2',
        's3fs>=0.4.2',
        'zarr>=2.12.0',
        'scikit-image>=0.19.3',
        'poisson-disc>=0.2.1',
    ],
    python_requires='>=3.9'
)
